#include <vector.h>
#include <math.h>
#include <fstream.h>
#include <iomanip.h>
#include "PacketDescriptor.h"
#include "FPL.h"

#define RRQ 1
#define RRP 2
#define RRC 4
#define duringRRP 3
#define duringDATA 5
#define DATA 6
#define ACKDATA 7
#define MAXLATENCY 500
#define LIGHTSPEED 3E8
#define maxPacketSize 2000
#define PULSELENGTH 1E-9
#define TRANSMITTER_RESET_TIME 0
//#define packetBitRate 1000000
#define maxInterferenceDistance 120
#define maxPowerContributionDistance 120

double distance(double x1, double y1, double x2, double  y2)
{
    
    return(sqrt(pow(x1-x2,2)+pow(y1-y2,2)));
} 
/*double min(double a, double b)
{
    if(a<b)
	return(a);
    else
	return(b);
}*/
/*
    insertInFPL
    Used to schedule a packet transmission in a terminal. It takes as inputs the expected transmission time, the packet duration and other 
    information related to packet/transmitter/connection identity.
    The output is either the effective packet transmission time, or an error if the transmission cannot be scheduled within MAXLATENCY seconds
*/

bool FPL::insertInFPL(int nodeID, int packetID, int destinationID, int connID, double expectedTxTime, int packetKind, double packetDuration, double *effectiveTxTime, bool tryPacket, bool SlottedAloha, double Slot_delay, double SlotDuration)
{
    //double actualTime=5;//DA CORREGGERE
    PacketDescriptor *pointer = startList;
    PacketDescriptor *temp;
    double nextTxTime;
    if(SlottedAloha)
      expectedTxTime+=Slot_delay;
    nextTxTime = expectedTxTime;
    //cout<<"Entrato in insertInFPL; nextTxTime: "<<nextTxTime<<"\n";
    while(pointer)
    {
    
    //Mi sposto all'istante previsto di trasmissione del pacchetto
	while(pointer && (pointer->_txTime <= expectedTxTime))
	{	
	    //We move to the expected packet transmission time...
	    //cout<<"incremento pointer\n";
	    temp=pointer;
	    pointer=pointer->next;
	    if(temp->_txTime + temp->_packetDuration >= expectedTxTime)
	    {    
	      //cout<<"condizione vera\n";
	      if(SlottedAloha)
		nextTxTime = temp->_txTime + SlotDuration;
	      else
		nextTxTime = temp->_txTime + temp->_packetDuration + TRANSMITTER_RESET_TIME;
	    }
	    else
	      ;
	      //cout<<"condizione falsa\n";
	    //cout<<"NextTxTime: "<<nextTxTime<<"\n";	  
	}
	if(pointer)
	{    
	    if(nextTxTime + packetDuration < pointer->_txTime)
	    {
		//We schedule the packet transmission in an intermediate position in the packet list
		PacketDescriptor *newPacket = new PacketDescriptor(nodeID,packetID, destinationID,connID,nextTxTime,packetKind,packetDuration,tryPacket);
		newPacket->next = pointer;
		if(pointer == startList)
		    startList = newPacket;
		else
		    temp->next = newPacket;
		*effectiveTxTime = nextTxTime;
		//cout<<"Pacchetto inserito in posizione intermedia\n";
		//cout<<"Effectivetime: "<< *effectiveTxTime<<endl;
		return(true);
	    }
	    else    
	    {
		//We move to the next position in the list
	      if(SlottedAloha)
		{
		  //cout<<"Here!"<<endl;
		  nextTxTime = pointer->_txTime + SlotDuration;
		}
	      else
		nextTxTime = pointer->_txTime + pointer->_packetDuration + TRANSMITTER_RESET_TIME;
	      temp = pointer;
	      pointer = pointer->next;
	    }
	}
	if(!pointer)
	{
	  //cout<<"NextTxTime: "<<nextTxTime<<"\n";	  
	  //We schedule the packet after the last scheduled packet 
	  PacketDescriptor *newPacket = new PacketDescriptor(nodeID,packetID, destinationID,connID,nextTxTime,packetKind,packetDuration,tryPacket);
	  newPacket->next=pointer;
	  temp->next=newPacket;
	  *effectiveTxTime = nextTxTime;
	  //cout<<"Pacchetto inserito in posizione finale\n";
	  //cout<<"Effectivetime: "<< *effectiveTxTime<<endl;
	  return(true);
	}
	if((packetKind!=DATA)&&(nextTxTime-expectedTxTime>MAXLATENCY))
	//non ci sono intervalli liberi in tempo utile: connessione rifiutata
	{
	    //Scheduling impossible within MAX LATENCY seconds: packet discarded
	    cout<<"Fallito"<<endl; 
	    return(false);
	}	    
    }
	
    //cout<<"la lista era vuota: aggiungo il nuovo pacchetto come primo\n";
    //We schedule the packet in the first position: the list was empty
    PacketDescriptor *newPacket = new PacketDescriptor(nodeID, packetID, destinationID, connID, nextTxTime, packetKind, packetDuration, tryPacket);
    newPacket->next=0;
    startList=newPacket;
    *effectiveTxTime=nextTxTime;
    //cout<<"Effectivetime: "<< *effectiveTxTime<<endl;
    return(true);		
}

void FPL::removeFromFPL(int connID, int nodeID)
{
    PacketDescriptor *pointer=startList;
    PacketDescriptor *previous=startList;
    while(pointer)
    {
	if((pointer->_connID==connID)&&(pointer->_nodeID==nodeID))
	{
	    if(pointer==startList)
	    {
		startList=pointer->next;
		previous=startList;
		delete pointer;
		pointer=previous->next;
	    }	
	    else
	    {
		previous->next=pointer->next;
		delete pointer;
		pointer=previous->next;
	    }
	}
	else
	{
	    previous=pointer;
	    pointer=pointer->next;
	}
    }
}	    

void FPL::removeFirstFromFPL()
{
    PacketDescriptor *pointer=startList;
    if(pointer->next)
	startList=pointer->next;
    else
	startList=0;
    delete pointer;
}	
//insertInFPLNoCheck: inserts packets in the FPL used for interference evaluation 
bool FPL::insertInFPLNoCheck(int nodeID, int packetID,int connID, double txTime, int packetKind, double packetDuration, double x, double y, double pTx)
{
    PacketDescriptor *pointer=startList;
    while(pointer)
	    {
		
		//pointer e' uguale a startList e c'e' spazio prima del primo pacchetto: inserisco il primo pacchetto in testa 
		
		if((pointer==startList)&&(txTime<pointer->_txTime))
		{
		    //pointer is equal to startlist, there is empty room at the beginning of the list: we insert the packet at list beginning
		    PacketDescriptor *newPacket = new PacketDescriptor(nodeID,packetID, -1,connID,txTime,packetKind,packetDuration,false,x,y,pTx);
		    newPacket->next=pointer->next;
		    startList=newPacket;
		    return(true);
		}
		if(pointer->next)
		//pointer non punta ancora all'ultimo pacchetto: tento di inserire il nuovo pacchetto
		{
		    //pointer is not set to the last packet: we try to insert the new packet
		    if((pointer->_txTime<=txTime) && ((pointer->next)->_txTime>txTime))
		    {
			PacketDescriptor *newPacket = new PacketDescriptor(nodeID,packetID, -1,connID,txTime,packetKind,packetDuration,false,x,y,pTx);
			newPacket->next=pointer->next;
			pointer->next=newPacket;
			return(true);
		    }
		    pointer=pointer->next;
		} 
		else
		//pointer punta all'ultimo pacchetto: aggiungo il nuovo pacchetto alla fine della lista
		{
		    //pointer is set to the last packet in the list: we insert the packet at list end 
		    PacketDescriptor *newPacket = new PacketDescriptor(nodeID,packetID, -1,connID,txTime,packetKind,packetDuration,false,x,y,pTx);
		    newPacket->next=0;
		    pointer->next=newPacket;
		    return(true);
		}
			    
	    }
	    //la lista era vuota: aggiungo il nuovo pacchetto come primo
	    //the list is empy: we insert the packet at the beginning
	    PacketDescriptor *newPacket = new PacketDescriptor(nodeID,packetID, -1,connID,txTime,packetKind,packetDuration,false, x,y,pTx);
	    newPacket->next=0;
	    startList=newPacket;
	    return(true);

}

//evaluateInterferencePower: evaluates interference power at time rxTime in position (x,y)

double FPL::evaluateInterferencePower(int nodeID, int packetID, double packetDuration, double rxTime, double x, double y, int txNodeID, int packetBitRate)
{
  ofstream intdebug;   
  PacketDescriptor *pointer=startList;
  //Definisco l'intervallo di ricerca nella lista
  //We define the time interval to be searched in the list, depending on area size, max packet size and transmission rate
  double startTime=rxTime-(double)maxPacketSize/packetBitRate-(double) maxInterferenceDistance/LIGHTSPEED;
  double interferringInterval;
  double totalInterferencePower=0;
  //intdebug.open("Interferencelog.txt", ios::app);
  //intdebug << "Node ["<< nodeID <<"]: received packet with ID " << packetID << " at " << setprecision(10) << rxTime << " from node[" << txNodeID << "]" << endl;
  //intdebug << "Node [" << nodeID << "]: Interferring packets list" << endl; 
  while(pointer)
    {
      if((pointer->_nodeID!=txNodeID) && (pointer->_nodeID!=nodeID))
	{
	  if(pointer->_txTime>=startTime)
	    if(pointer->_txTime<=rxTime)
	      {
		//Il pacchetto puo' creare interferenza: controllo la distanza tra trasmettitore e ricevitore
		//The packet may cause interference: we check distance between transmitter and receiver in (x,y)
		double dist=distance(pointer->_x,pointer->_y,x,y);
		if(dist<=maxInterferenceDistance)
		  {
		    if(dist<1)
		      dist=1;
		    
		    //Transmitter is close enough to interfere with receiver: we check if the potentially interfering packet collides with 
		    //the useful packet
		    double propagationDelay=dist/LIGHTSPEED;
		    if(pointer->_txTime+propagationDelay>=rxTime-packetDuration)
		      interferringInterval=min(rxTime-(pointer->_txTime+propagationDelay),pointer->_packetDuration);
		    else
		      interferringInterval=min(pointer->_txTime+propagationDelay+pointer->_packetDuration-(rxTime-packetDuration),packetDuration);

		    if(interferringInterval>=0)
		      {
			//The two packets collide: we evaluate the contribution to the interference power
			totalInterferencePower+=(pointer->_pTx)*(interferringInterval/packetDuration)*((double) PULSELENGTH*packetBitRate)/pow(dist,2);
			//totalInterferencePower+=(pointer->_pTx)*(interferringInterval/packetDuration)/pow(dist,2);
			//intdebug << pointer->_packetID << " " << pointer->_nodeID << " " << pointer->_packetKind << " " << pointer->_txTime << " " << pointer->_txTime + pointer->_packetDuration << " " << interferringInterval/packetDuration << " " << setprecision(10) << (pointer->_pTx) * (interferringInterval / packetDuration) * ((double) PULSELENGTH * packetBitRate) / pow(dist,2) <<  " " << totalInterferencePower << endl;
	  
	
		//			    printList();
		      }
		  }
	      }
	    else
	      pointer=0;
	}
      else
	{
	  //	    cout<<"Nodorx: "<<nodeID<<" evitata autointerferenza"<<endl;
	  ;
	}
      if(pointer)
	pointer=pointer->next;
    }
  //intdebug << totalInterferencePower << endl;
  //intdebug.close();
    return(totalInterferencePower);
}

void FPL::evaluateCollidingPackets(int nodeID, int packetID, int packetSize, double packetRxTime, double x, double y, int packetBitRate, int txNodeID, int *interferringPackets)
{
  ofstream intdebug;   

  //Definisco l'intervallo di ricerca nella lista
  //We define the time interval to be searched in the list, depending on area size, max packet size and transmission rate
  double startTime=packetRxTime-(double)maxPacketSize/packetBitRate-(double) maxInterferenceDistance/LIGHTSPEED;
  double interferringInterval;
  //double totalInterferencePower=0;
  //intdebug.open("Interferencelog.txt", ios::app);
  //intdebug << "Node ["<< nodeID <<"]: received packet with ID " << packetID << " at " << setprecision(10) << rxTime << " from node[" << txNodeID << "]" << endl;
  //intdebug << "Node [" << nodeID << "]: Interferring packets list" << endl; 
  for(int bitNumber = 1; bitNumber <= packetSize; bitNumber++)
    {
      double bitStartTime = packetRxTime - (double) bitNumber / packetBitRate;
      double bitEndTime = packetRxTime - (double) (bitNumber - 1) / packetBitRate;
      double bitInterval = bitEndTime - bitStartTime;
      int Packets = 0;
      PacketDescriptor *pointer = startList;
      while(pointer)
	{
	  if((pointer->_nodeID!=txNodeID) && (pointer->_nodeID!=nodeID))
	    {
	      if(pointer->_txTime>=bitStartTime)
		if(pointer->_txTime<=bitEndTime)
		  {
		    //Il pacchetto puo' creare interferenza: controllo la distanza tra trasmettitore e ricevitore
		    //The packet may cause interference: we check distance between transmitter and receiver in (x,y)
		    double dist=distance(pointer->_x,pointer->_y,x,y);
		    if(dist<=maxInterferenceDistance)
		      {
			//			if(dist<1)
			//dist=1;
			
			//Transmitter is close enough to interfere with receiver: we check if the potentially interfering packet collides with 
			//the useful packet
			double propagationDelay=dist/LIGHTSPEED;
			if(pointer->_txTime + propagationDelay >= bitStartTime)
			  interferringInterval=min(bitEndTime - (pointer->_txTime+propagationDelay), pointer->_packetDuration);
			else
			  interferringInterval=min(pointer->_txTime+propagationDelay+pointer->_packetDuration-bitStartTime, bitInterval);
			
			if(interferringInterval>=0)
			  {
			    //The two packets collide in this bit: we evaluate the contribution to the interference
			    Packets++;
			    
			  }
		      }
		  }
		else
		  pointer=0;
	    }
	  else
	    {
	      //	    cout<<"Nodorx: "<<nodeID<<" evitata autointerferenza"<<endl;
	      ;
	    }
	  if(pointer)
	    pointer=pointer->next;
	}
      interferringPackets[packetSize-bitNumber] = Packets;
      //cout << Packets << interferringPackets[packetSize - bitNumber] << endl;
    }
  //intdebug << totalInterferencePower << endl;
  //intdebug.close();
  //return(totalInterferencePower);
}

double FPL::evaluateTotalPower(double rxTime, double x, double y, double Integration_time, int packetBitRate)
{
    PacketDescriptor *pointer=startList;
    //Definisco l'intervallo di ricerca nella lista
    double startTime=rxTime-(double)maxPacketSize/packetBitRate-(double) maxPowerContributionDistance/LIGHTSPEED - Integration_time;
    double interferringInterval;
    double totalPower=0;
    cout <<"Start time= "<<startTime<<" End Time= "<<rxTime<<endl;
    //cout << "x= "<< x <<" y= "<< y<<endl;
    //    cout<<"Nodorx: "<<nodeID<<" ricevuto pacchetto con ID "<<packetID<<" in "<<rxTime<<" dal nodo "<<txNodeID<<endl;
    while(pointer)
    {
	    //cout<<"pointer ok"<<endl;
	    if(pointer->_txTime>=startTime)
		if(pointer->_txTime<=rxTime)
		{
		    double dist=distance(pointer->_x,pointer->_y,x,y);
		    //cout<<"Evaluated distance"<<endl;
		    if(dist<=maxPowerContributionDistance)
		    {
			if(dist<1)
			    dist=1;
			double propagationDelay=dist/LIGHTSPEED;
			if(pointer->_txTime+propagationDelay+pointer->_packetDuration>=rxTime)
			{    
			    totalPower+=(double) (pointer->_pTx)/pow(dist,2);
			    //cout<<"Total power: "<<totalPower<<endl;
			}
//			    cout<<"Nodorx: "<<nodeID<<" interferito da pacchetto con ID "<<pointer->_packetID<<" trasmesso in "<< pointer->_txTime <<" con potenza "<<pointer->_pTx<<" dal nodo "<<pointer->_nodeID<<" a distanza "<<dist<<"; coefficiente di sovrapposizione: "<<interferringInterval/packetDuration<<endl; 
//			    cout<<"Nodorx: "<<nodeID<<" subisce interferenza pari a "<<(pointer->_pTx)*(interferringInterval/packetDuration)*((double) PULSELENGTH*packetBitRate)/pow(dist,2)<<endl;		    
//			    printList();		
		    }
		}
		else
		    pointer=0;
	    if(pointer)
		pointer=pointer->next;
    }
    return(totalPower);
}

void FPL::refreshList(double actualTime, double preservingInterval)
{
    PacketDescriptor *pointer=startList;
    while(pointer)
    {
	if(pointer->_txTime<actualTime-preservingInterval)
	{
	    removeFirstFromFPL();
	    pointer=startList;
	}
	else
	    break;
    }
}

void FPL::printList()
{
  PacketDescriptor *pointer=startList;
  while(pointer)
    {
      cout << setprecision(10) << "Packet: " << pointer->_packetID << " Node: " << pointer->_nodeID <<" TXTime: "<< pointer->_txTime << " EndTime: " << pointer->_txTime + pointer->_packetDuration << " Power: " << pointer->_pTx << " packetKind: " << pointer->_packetKind << endl; 
      pointer = pointer->next;
    }
  cout << endl;
}
