#ifndef _FPL
#define _FPL
/*
    FPL.h
    Used to track all packet transmissions in the network:
    
    - Each terminal maintains a local packet list, which is used to schedule terminal packet transmissions
    
    - Connectivity module uses a Future Packet List to evaluate the interference caused on each received packet
*/

class FPL{

    public:
    
	FPL():startList(0),endList(0){}
	//insertInFPL: Used by each terminal to schedule its own packets
	bool insertInFPL(int nodeID, int packetID, int destinationID, int connID, double expectedTxTime, int packetKind, double packetDuration, double *effectiveTxTime, bool trypacket, bool SlottedAloha, double Slot_delay, double SlotDuration);
	//removeFromFPL: Used to remove all packets of a connection from the Future Packet List
	void removeFromFPL(int connID, int nodeID);
	void removeFirstFromFPL();
	//refreshList: removes from FPL pointers of past packets which cannot cause interference anymore
	void refreshList(double actualTime, double preservingInterval);
	void printList();
	//insertInFPL: Used by each terminal to record its transmission in the global Future Packet List
	bool insertInFPLNoCheck(int nodeID, int packetID, int connID, double txTime, int packetKind, double packetDuration, double x, double y, double pTx);
	//evaluateInterferencePower: evaluates the amount of interference at time rxTime in the point (x,y)  
	double evaluateInterferencePower(int nodeID, int packetID, double packetDuration, double rxTime, double x, double y, int txNodeID, int packetRate);
	void evaluateCollidingPackets(int nodeID, int packetID, int packetSize, double packetRxTime, double x, double y, int packetBitRate, int txNodeID, int *interferringPackets);
	double evaluateTotalPower(double rxTime, double x, double y, double Integration_time, int packetRate);
	PacketDescriptor *startList;
	PacketDescriptor *endList;

};
#endif
